/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ServiceWorkerContextData.h"
#include <wtf/CrossThreadCopier.h>

namespace WebCore {

ServiceWorkerContextData ServiceWorkerContextData::isolatedCopy() const &
{
    return {
        jobDataIdentifier,
        registration.isolatedCopy(),
        serviceWorkerIdentifier,
        script.isolatedCopy(),
        certificateInfo.isolatedCopy(),
        contentSecurityPolicy.isolatedCopy(),
        crossOriginEmbedderPolicy.isolatedCopy(),
        referrerPolicy.isolatedCopy(),
        scriptURL.isolatedCopy(),
        workerType,
        loadedFromDisk,
        lastNavigationWasAppInitiated,
        crossThreadCopy(scriptResourceMap),
        serviceWorkerPageIdentifier,
        crossThreadCopy(navigationPreloadState),
        crossThreadCopy(routes)
    };
}

ServiceWorkerContextData ServiceWorkerContextData::isolatedCopy() &&
{
    return {
        jobDataIdentifier,
        WTF::move(registration).isolatedCopy(),
        serviceWorkerIdentifier,
        WTF::move(script).isolatedCopy(),
        WTF::move(certificateInfo).isolatedCopy(),
        WTF::move(contentSecurityPolicy).isolatedCopy(),
        WTF::move(crossOriginEmbedderPolicy).isolatedCopy(),
        WTF::move(referrerPolicy).isolatedCopy(),
        WTF::move(scriptURL).isolatedCopy(),
        workerType,
        loadedFromDisk,
        lastNavigationWasAppInitiated,
        crossThreadCopy(WTF::move(scriptResourceMap)),
        serviceWorkerPageIdentifier,
        crossThreadCopy(WTF::move(navigationPreloadState)),
        crossThreadCopy(WTF::move(routes))
    };
}

ServiceWorkerContextData ServiceWorkerContextData::copy() const
{
    return {
        jobDataIdentifier,
        registration,
        serviceWorkerIdentifier,
        script,
        certificateInfo,
        contentSecurityPolicy,
        crossOriginEmbedderPolicy,
        referrerPolicy,
        scriptURL,
        workerType,
        loadedFromDisk,
        lastNavigationWasAppInitiated,
        scriptResourceMap,
        serviceWorkerPageIdentifier,
        navigationPreloadState,
        map(routes, [](auto& route) { return route.copy(); })
    };
}

} // namespace WebCore
