/*
 * Copyright (C) 2021-2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

// https://bugs.webkit.org/show_bug.cgi?id=232548 These shouldn't need to be here.
typedef [EnforceRange] unsigned long GPUSize32;
typedef [EnforceRange] unsigned long long GPUSize64;
typedef [EnforceRange] unsigned long GPUIndex32;
typedef [EnforceRange] long GPUSignedOffset32;

[
    EnabledBySetting=WebGPUEnabled
]
interface mixin GPURenderEncoderBase {
    undefined setPipeline(GPURenderPipeline pipeline);

    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=240219, last two parameters should have default value of 0
    undefined setIndexBuffer(GPUBuffer buffer, GPUIndexFormat indexFormat, optional GPUSize64 offset, optional GPUSize64 size);
    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=240219, last two parameters should have default value of 0
    undefined setVertexBuffer(GPUIndex32 slot, GPUBuffer? buffer, optional GPUSize64 offset, optional GPUSize64 size);

    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=240219, last two parameters should have default value of 0, first 1
    undefined draw(GPUSize32 vertexCount, optional GPUSize32 instanceCount,
        optional GPUSize32 firstVertex, optional GPUSize32 firstInstance);
    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=240219, last three parameters should have default value of 0, first 1
    undefined drawIndexed(GPUSize32 indexCount, optional GPUSize32 instanceCount,
        optional GPUSize32 firstIndex,
        optional GPUSignedOffset32 baseVertex,
        optional GPUSize32 firstInstance);

    undefined drawIndirect(GPUBuffer indirectBuffer, GPUSize64 indirectOffset);
    undefined drawIndexedIndirect(GPUBuffer indirectBuffer, GPUSize64 indirectOffset);
};
