/*
 * Copyright (C) 2019-2025 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "Supplementable.h"
#include <wtf/CheckedRef.h>
#include <wtf/TZoneMalloc.h>
#include <wtf/WeakPtr.h>

namespace WebCore {

class DeferredPromise;
class Document;
class Navigator;
enum class IsLoggedIn : uint8_t;

class NavigatorLoginStatus final : public Supplement<Navigator>, public CanMakeWeakPtr<NavigatorLoginStatus> {
    WTF_MAKE_TZONE_ALLOCATED(NavigatorLoginStatus);
public:
    explicit NavigatorLoginStatus(Navigator& navigator)
        : m_navigator(navigator)
    {
    }
    static void setStatus(Navigator&, IsLoggedIn, Ref<DeferredPromise>&&);
    static void isLoggedIn(Navigator&, Ref<DeferredPromise>&&);

private:
    void setStatus(IsLoggedIn, Ref<DeferredPromise>&&);
    void isLoggedIn(Ref<DeferredPromise>&&);

    static NavigatorLoginStatus* from(Navigator&);
    static ASCIILiteral supplementName() { return "NavigatorLoginStatus"_s; }
    bool isNavigatorLoginStatus() const final { return true; }
    bool hasSameOrigin() const;

    const CheckedRef<Navigator> m_navigator;
};

} // namespace WebCore

SPECIALIZE_TYPE_TRAITS_BEGIN(WebCore::NavigatorLoginStatus)
    static bool isType(const WebCore::SupplementBase& supplement) { return supplement.isNavigatorLoginStatus(); }
SPECIALIZE_TYPE_TRAITS_END()
