/*
 * Copyright (C) 2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <WebCore/InlineDisplayBox.h>
#include <WebCore/InlineDisplayLine.h>
#include <wtf/Vector.h>

namespace WebCore {
namespace InlineDisplay {

using Boxes = Vector<Box, 2>;
using Lines = Vector<Line, 1>;

struct Content {
    WTF_DEPRECATED_MAKE_STRUCT_FAST_ALLOCATED(Content);

    void clear();
    void set(Content&& newContent);
    void append(Content&& newContent);
    void insert(Content&& newContent, size_t lineIndex, size_t boxIndex);
    void remove(size_t firstLineIndex, size_t numberOfLines, size_t firstBoxIndex, size_t numberOfBoxes);

    std::optional<Line::Ellipsis> lineEllipsis(size_t) const;
    void setLineEllipsis(size_t line, Line::Ellipsis&&);

    void moveLineInBlockDirection(size_t, float offset);
    void shrinkLineInBlockDirection(size_t, float delta);

    Lines lines;
    Boxes boxes;

private:
    using LineEllipses = Vector<std::optional<Line::Ellipsis>>;
    std::unique_ptr<LineEllipses> lineEllipses;
};

}
}
