/*
 * Copyright (C) 2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <WebCore/EventTarget.h>
#include <WebCore/WebSocketFrame.h>
#include <wtf/Forward.h>
#include <wtf/ObjectIdentifier.h>

namespace WebCore {

class Document;
class WeakPtrImplWithEventTargetData;
class ResourceRequest;
class ResourceResponse;
class WebSocketChannel;
class WebSocketChannelInspector;

using WebSocketChannelIdentifier = AtomicObjectIdentifier<WebSocketChannel>;

class WEBCORE_EXPORT WebSocketChannelInspector {
public:
    explicit WebSocketChannelInspector(Document&);
    ~WebSocketChannelInspector();

    void didCreateWebSocket(const URL&) const;
    void willSendWebSocketHandshakeRequest(const ResourceRequest&) const;
    void didReceiveWebSocketHandshakeResponse(const ResourceResponse&) const;
    void didCloseWebSocket() const;
    void didReceiveWebSocketFrame(const WebSocketFrame&) const;
    void didSendWebSocketFrame(const WebSocketFrame&) const;
    void didReceiveWebSocketFrameError(const String& errorMessage) const;
    
    WebSocketChannelIdentifier progressIdentifier() const { return m_progressIdentifier; }

    static WebSocketFrame createFrame(std::span<const uint8_t> data, WebSocketFrame::OpCode);

private:
    WeakPtr<Document, WeakPtrImplWithEventTargetData> m_document;
    WebSocketChannelIdentifier m_progressIdentifier;
};

} // namespace WebCore
