/*
 * Copyright (C) 2023 Igalia S.L
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * aint with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

[
 Conditional=WEB_CODECS
 ] enum OpusBitstreamFormat {
    "opus",
    "ogg",
};

typedef [EnforceRange] unsigned long WebCodecsAudioOpusEncoderConfigParameter;
typedef [EnforceRange] unsigned long long WebCodecsAudioOpusEncoderConfigFrameDuration;

[
    Conditional=WEB_CODECS,
    JSGenerateToJSObject,
    JSGenerateToNativeObject,
    LegacyNativeDictionaryRequiredInterfaceNullability,
] dictionary OpusEncoderConfig {
    OpusBitstreamFormat format = "opus";
    WebCodecsAudioOpusEncoderConfigFrameDuration frameDuration = 20000;
    WebCodecsAudioOpusEncoderConfigParameter complexity;
    WebCodecsAudioOpusEncoderConfigParameter packetlossperc = 0;
    boolean useinbandfec = false;
    boolean usedtx = false;
};
