# Copyright (C) 2011 Apple Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.

require "config"
require "digest/sha1"
require "pathname"

#
# dirHash(directory, regexp) -> SHA1 hexdigest
#
# Returns a hash of all files in the given directory that fit the given
# pattern.
#

def dirHash(directory, regexp)
    directory = Pathname.new(directory)
    contents = ""
    Dir.foreach(directory) {
        | entry |
        if entry =~ regexp
            contents += IO::read(directory + entry)
        end
    }
    return Digest::SHA1.hexdigest(contents)
end

#
# fileListHash(fileList) -> SHA1 hexdigest
#
# Returns a hash of all files in the list.
#

def fileListHash(fileList)
    contents = ""
    fileList.each {
        | fileName |
        contents += IO::read(fileName)
    }
    return Digest::SHA1.hexdigest(contents)
end

#
# selfHash -> SHA1 hexdigest
#
# Returns a hash of the offlineasm source code. This allows dependency
# tracking for not just changes in input, but also changes in the assembler
# itself.
#

def selfHash
    dirHash(Pathname.new(__FILE__).dirname, /\.rb$/)
end

