/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// https://w3c.github.io/IndexedDB/#idbcursor
[
    EnabledBySetting=IndexedDBAPIEnabled,
    JSCustomMarkFunction,
    SkipVTableValidation,
    Exposed=(Window,Worker)
] interface IDBCursor {
    readonly attribute (IDBObjectStore or IDBIndex) source;
    readonly attribute IDBCursorDirection direction;
    [CustomGetter] readonly attribute any key;
    [CustomGetter] readonly attribute any primaryKey;
    // FIXME: `request` should not be nullable.
    [SameObject] readonly attribute IDBRequest? request;

    undefined advance([EnforceRange] unsigned long count);
    [CallWith=CurrentGlobalObject, ImplementedAs=continueFunction] undefined continue(optional any key);
    [CallWith=CurrentGlobalObject] undefined continuePrimaryKey(any key, any primaryKey);

    [NewObject, CallWith=CurrentGlobalObject] IDBRequest update(any value);
    [NewObject, ImplementedAs=deleteFunction] IDBRequest delete();
};
