/*
 * Copyright (C) 2011, 2012 Google Inc. All rights reserved.
 * Copyright (C) 2024 Samuel Weinig <sam@webkit.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <WebCore/CSSCalcTree.h>
#include <wtf/RefCounted.h>
#include <wtf/Ref.h>

namespace WebCore {

namespace Style {
namespace Calculation {
class Value;
}
}

namespace CSS {
struct PropertyParserState;
struct Range;
enum class Category : uint8_t;
}

class CSSCalcSymbolTable;
class CSSCalcSymbolsAllowed;
class CSSParserTokenRange;
class CSSToLengthConversionData;
class RenderStyle;

struct CSSParserContext;
struct CSSPropertyParserOptions;
struct NoConversionDataRequiredToken;

enum CSSValueID : uint16_t;

enum class CSSUnitType : uint8_t;

namespace CSSCalc {

// Boxes a `CSSCalc::Tree` along with the `Style::Calculation::Category` and `CSS::Range` used to construct it.
class Value final : public RefCounted<Value> {
public:
    static RefPtr<Value> parse(CSSParserTokenRange&, CSS::PropertyParserState&, CSS::Category, CSS::Range, CSSCalcSymbolsAllowed, CSSPropertyParserOptions);

    static Ref<Value> create(const Style::Calculation::Value&, const RenderStyle&);
    static Ref<Value> create(CSS::Category, CSS::Range, CSSCalc::Tree&&);

    ~Value();

    // Creates a copy of the CSSCalc::Tree with non-canonical dimensions and any symbols present in the provided symbol table resolved.
    Ref<Value> copySimplified(const CSSToLengthConversionData&) const;
    Ref<Value> copySimplified(const CSSToLengthConversionData&, const CSSCalcSymbolTable&) const;
    Ref<Value> copySimplified(NoConversionDataRequiredToken) const;
    Ref<Value> copySimplified(NoConversionDataRequiredToken, const CSSCalcSymbolTable&) const;

    CSS::Category category() const { return m_category; }
    CSS::Range range() const { return m_range; }

    CSSUnitType primitiveType() const;

    // Returns whether the CSSCalc::Tree requires `CSSToLengthConversionData` to fully resolve.
    bool requiresConversionData() const { return m_tree.requiresConversionData; };

    double doubleValue(const CSSToLengthConversionData&) const;
    double doubleValue(const CSSToLengthConversionData&, const CSSCalcSymbolTable&) const;
    double doubleValue(NoConversionDataRequiredToken) const;
    double doubleValue(NoConversionDataRequiredToken, const CSSCalcSymbolTable&) const;
    double doubleValueDeprecated() const;

    double computeLengthPx(const CSSToLengthConversionData&) const;
    double computeLengthPx(const CSSToLengthConversionData&, const CSSCalcSymbolTable&) const;

    Ref<Style::Calculation::Value> createCalculationValue(NoConversionDataRequiredToken) const;
    Ref<Style::Calculation::Value> createCalculationValue(NoConversionDataRequiredToken, const CSSCalcSymbolTable&) const;
    Ref<Style::Calculation::Value> createCalculationValue(const CSSToLengthConversionData&) const;
    Ref<Style::Calculation::Value> createCalculationValue(const CSSToLengthConversionData&, const CSSCalcSymbolTable&) const;

    void collectComputedStyleDependencies(ComputedStyleDependencies&) const;

    String cssText(const CSS::SerializationContext&) const;
    bool equals(const Value&) const;

    void dump(TextStream&) const;

    const CSSCalc::Tree& tree() const { return m_tree; }

private:
    explicit Value(CSS::Category, CSS::Range, CSSCalc::Tree&&);

    double clampToPermittedRange(double) const;

    CSS::Category m_category;
    CSS::Range m_range;
    Tree m_tree;
};

TextStream& operator<<(TextStream&, const Value&);

} // namespace CSSCalc
} // namespace WebCore
