/*
 * Copyright (C) 2025 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "RunLoopObserver.h"

#if USE(GLIB_EVENT_LOOP)
namespace WebCore {

void RunLoopObserver::schedule(PlatformRunLoop runLoop, OptionSet<Activity> activities)
{
    Locker lock { m_runLoopObserverLock };
    if (m_runLoopObserver)
        return;

    if (!runLoop)
        runLoop = RunLoop::currentSingleton();

    m_runLoopObserver = ActivityObserver::create(runLoop.releaseNonNull(), isRepeating(), static_cast<uint8_t>(m_order), activities, [this]() {
        runLoopObserverFired();
    });

    m_runLoopObserver->start();
}

void RunLoopObserver::invalidate()
{
    Locker lock { m_runLoopObserverLock };
    if (m_runLoopObserver) {
        m_runLoopObserver->stop();
        m_runLoopObserver = nullptr;
    }
}

bool RunLoopObserver::isScheduled() const
{
    Locker lock { m_runLoopObserverLock };
    return !!m_runLoopObserver;
}

} // namespace WebCore

#endif // USE(GLIB_EVENT_LOOP)
