/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

.tree-outline.dom .badge {
    /* Using sans-serif San Francisco font here creates a badge 1px taller than the selected
    area. Use the same monospace font as the rest of the DOM tree outline. */

    font-size: 8px;
    vertical-align: 1px;
    color: var(--text-color);
    background: hsla(0, 0%, 95%, 0.8);
    border: 1px solid hsla(0, 0%, var(--foreground-lightness), 0.1);
    border-radius: 3px;
    margin-left: 3px;
    padding: 0 2px;
    box-shadow: 0 0 1px hsla(0, 0%, 0%, 0.5);
}

.tree-outline.dom .badge.activated {
    background: var(--glyph-color-active);
    color: white;
}

body:not(.window-inactive, .window-docked-inactive) .tree-outline.dom:focus-within li.selected .badge {
    /* Override `li.selected * {color: inherited}` from DOMTreeOutline.css */
    color: var(--text-color);
}

.event-badge-popover-content, .slot-filled-badge-popover-content {
    overflow: hidden;
}

.event-badge-popover-content > .details-section.event-listeners {
    max-height: calc(100vh - var(--tab-bar-height) - var(--console-prompt-min-height));
    overflow: auto;
    will-change: z-index; /* Workaround for <rdar://125601926> - Content renders on top of the scroll bar */
}

.event-badge-popover-content > .details-section.event-listeners .details-section {
    border-inline: 1px solid var(--border-color);
}

.slot-filled-badge-popover-content > table {
    border-collapse: collapse;
}

.slot-filled-badge-popover-content > table > tr + tr > td {
    padding-top: 2px;
}

.slot-filled-badge-popover-content > table > tr > td:first-child {
    font-family: -webkit-system-font, sans-serif;
    font-size: 11px;
    font-variant-numeric: tabular-nums;
    text-align: end;
    vertical-align: top;
    color: var(--console-secondary-text-color);
}

@media (prefers-color-scheme: dark) {
    .tree-outline.dom .badge {
        background: hsla(0, 0%, 30%, 0.8);
        border-color: hsla(0, 0%, 100%, 0.1);
    }
}
