/*
 * Copyright (C) 2020 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "PlatformScreen.h"

#if PLATFORM(COCOA) || PLATFORM(GTK) || (PLATFORM(WPE) && ENABLE(WPE_PLATFORM))

#include "ScreenProperties.h"
#include <wtf/NeverDestroyed.h>

namespace WebCore {

static ScreenProperties& screenProperties()
{
    static NeverDestroyed<ScreenProperties> screenProperties;
    return screenProperties;
}

const ScreenProperties& getScreenProperties()
{
    return screenProperties();
}

PlatformDisplayID primaryScreenDisplayID()
{
    return screenProperties().primaryDisplayID;
}

void setScreenProperties(const ScreenProperties& properties)
{
    screenProperties() = properties;
}

const ScreenData* screenData(PlatformDisplayID screenDisplayID)
{
    if (screenProperties().screenDataMap.isEmpty())
        return nullptr;

    // Return property of the first screen if the screen is not found in the map.
    if (auto displayID = screenDisplayID ? screenDisplayID : primaryScreenDisplayID()) {
        auto properties = screenProperties().screenDataMap.find(displayID);
        if (properties != screenProperties().screenDataMap.end())
            return &properties->value;
    }

    // Last resort: use the first item in the screen list.
    return &screenProperties().screenDataMap.begin()->value;
}

#if HAVE(SUPPORT_HDR_DISPLAY)
void setScreenContentsFormatsForTesting(OptionSet<ContentsFormat> contentsFormats)
{
    screenProperties().screenContentsFormatsForTesting = contentsFormats;
}

OptionSet<ContentsFormat> screenContentsFormatsForTesting()
{
    return screenProperties().screenContentsFormatsForTesting;
}
#endif

} // namespace WebCore

#endif // PLATFORM(COCOA) || PLATFORM(GTK) || (PLATFORM(WPE) && ENABLE(WPE_PLATFORM))
